#ifdef __linux__
#define _XOPEN_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#ifdef __linux__
#include <sys/select.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/if.h>
#include <linux/if_tun.h>
#endif

#define VECTOR_NOINTR 0xffff

#define VECTOR_SERIAL_RECV 18
#define VECTOR_SERIAL_SEND 19

#define IO_OFFSET 0x20

#if 0
#define MEM8(addr) (*(volatile unsigned char *)(addr))
#else
#define MEM8(addr) (memory[addr])
#endif
#define IO8(addr)  MEM8((addr) + IO_OFFSET)

#define USART_UCSR0A MEM8(0xc0)

#define USART_UCSR0A_UDRE0 (1<<5) /* Хåեǥå */
#define USART_UCSR0A_TXC0  (1<<6) /* λǥå */
#define USART_UCSR0A_RXC0  (1<<7) /* Хåե˥ǡͭǥå */

#define USART_UCSR0B MEM8(0xc1)
#define USART_UCSR0B_UCSZ02 (1<<2)
#define USART_UCSR0B_TXEN0  (1<<3)
#define USART_UCSR0B_RXEN0  (1<<4)
#define USART_UCSR0B_UDRIE0 (1<<5)
#define USART_UCSR0B_TXCIE0 (1<<6)
#define USART_UCSR0B_RXCIE0 (1<<7)

#define USART_UCSR0C MEM8(0xc2)
#define USART_UCSR0C_UCSZ00      (1<<1)
#define USART_UCSR0C_UCSZ01      (1<<2)
#define USART_UCSR0C_STOPBIT_1   (0<<3)
#define USART_UCSR0C_STOPBIT_2   (1<<3)
#define USART_UCSR0C_PARITY_NONE (0<<4)
#define USART_UCSR0C_PARITY_EVEN (2<<4)
#define USART_UCSR0C_PARITY_ODD  (3<<4)
#define USART_UCSR0C_MODE_ASYNC  (0<<6)
#define USART_UCSR0C_MODE_SYNC   (1<<6)
#define USART_UCSR0C_MODE_SPI    (3<<6)

#define USART_UBRR0L MEM8(0xc4)
#define USART_UBRR0H MEM8(0xc5)

#define USART_UDR0   MEM8(0xc6)

#define PIN_DDRD     IO8(0x0a)
#define PIN_PORTD    IO8(0x0b)
#define PIN_DDD0     (1<<0)
#define PIN_PORTD0   (1<<0)

#define MCUCR_IO    IO8(0x35)
#define MCUCR_IVCE  (1<<0)
#define MCUCR_IVSEL (1<<1)

static int serial_fd = -1;

static int search_freepty()
{
  int fd;

  /* use posix_openpt for same code with FreeBSD and Linux */

  do {
    char *devname;
    /*
     * use posix_openpt() for same code with FreeBSD and Linux.
     * old Linux are need to use direct open to /dev/ptmx.
     */
#if 1
    fd = posix_openpt(O_RDWR /*| O_NOCTTY*/);
#else /* for old Linux */
    fd = open ("/dev/ptmx", O_RDWR);
#endif
    if (fd < 0) break;

    if (grantpt(fd) < 0) break;
    if (unlockpt(fd) < 0) break;

    devname = ptsname(fd);
    if (devname == NULL) break;
    fprintf(stderr, "connect to %s\n", devname);
    return fd;
  } while (0);
  if (fd >= 0) {
    close(fd);
  }

  fprintf(stderr, "Free pty not found.\n");
  return -1;
}

static void serial_init(unsigned char *memory)
{
  USART_UCSR0A = 0;
  USART_UCSR0B = 0;
  USART_UCSR0C = 0;
  USART_UBRR0L = 0;
  USART_UBRR0H = 0;
  USART_UDR0   = 0;

  serial_fd = search_freepty();
}

static int is_send_enable(int fd)
{
  fd_set fds;
  struct timeval tm;
  int ret;

  if (fd < 0) return 0;

  FD_ZERO(&fds);
  FD_SET(fd, &fds);
  tm.tv_sec  = 0;
  tm.tv_usec = 0;

  ret = select(fd + 1, NULL, &fds, NULL, &tm);
  if (ret > 0) {
    if (FD_ISSET(fd, &fds)) {
      return 1;
    }
  }

  return 0;
}

static int is_recv_enable(int fd)
{
  fd_set fds;
  struct timeval tm;
  int ret;

  if (fd < 0) return 0;

  FD_ZERO(&fds);
  FD_SET(fd, &fds);
  tm.tv_sec  = 0;
  tm.tv_usec = 0;

  ret = select(fd + 1, &fds, NULL, NULL, &tm);
  if (ret > 0) {
    if (FD_ISSET(fd, &fds)) {
      return 1;
    }
  }

  return 0;
}

static void send_byte(int fd, unsigned char c)
{
  if (write (fd, &c, 1) != 1)
    ;
  fsync(fd);
}

static unsigned char recv_byte(int fd)
{
  unsigned char c;
  int size;
  size = read (fd, &c, 1);
  if (size == 1) {
    return c;
  }
  return 0;
}

static int serial_run(unsigned char *memory)
{
  static int send_intr = 0, recv_intr = 0;
  int vector = VECTOR_NOINTR;

  if (USART_UCSR0B & USART_UCSR0B_TXEN0) {
    if (!(USART_UCSR0A & USART_UCSR0A_TXC0) &&
	!(USART_UCSR0A & USART_UCSR0A_RXC0)) {
      if (is_send_enable(serial_fd)) {
	send_byte(serial_fd, USART_UDR0);
	USART_UCSR0A |= USART_UCSR0A_TXC0 | USART_UCSR0A_UDRE0;
      }
    }
  }
  send_intr =
    (USART_UCSR0A & USART_UCSR0A_TXC0) && (USART_UCSR0B & USART_UCSR0B_TXCIE0);

  if (USART_UCSR0B & USART_UCSR0B_RXEN0) {
    if (!(USART_UCSR0A & USART_UCSR0A_RXC0) &&
	(USART_UCSR0A & USART_UCSR0A_TXC0)) {
      if (is_recv_enable(serial_fd)) {
	USART_UDR0 = recv_byte(serial_fd);
	USART_UCSR0A |= USART_UCSR0A_RXC0;
      }
    }
  }
  recv_intr =
    (USART_UCSR0A & USART_UCSR0A_RXC0) && (USART_UCSR0B & USART_UCSR0B_UDRIE0);

  if (send_intr) vector = VECTOR_SERIAL_SEND;
  if (recv_intr) vector = VECTOR_SERIAL_RECV;

  return vector;
}

void sim_device_init(unsigned char *memory)
{
  static int initialized = 0;

  if (initialized)
    return;

  serial_init(memory);

  initialized++;
}

int sim_device_sleep(unsigned char *memory)
{
  fd_set fds;
  int ret, fd = -1;

  if (fd < serial_fd) fd = serial_fd;
  if (fd < 0) {
    fprintf(stderr, "not found opened device.\n");
  }

  FD_ZERO(&fds);
  if (serial_fd >= 0) {
    FD_SET(serial_fd, &fds);
  }

  ret = select(fd + 1, &fds, NULL, NULL, NULL);
  if (ret > 0) {
    if (serial_fd >= 0)
      if (FD_ISSET(serial_fd, &fds))
	return 1;
  }

  return 0;
}

int sim_device_run(unsigned char *memory)
{
  int vec, vector = VECTOR_NOINTR;

  vec    = serial_run(memory);
  vector = (vec < vector) ? vec : vector;

  return (vector == VECTOR_NOINTR) ? 0 : vector;
}
